import re
import typing


def auto_group_list_by_pkeys(
    primary_key_names: tuple, list_of_dicts: typing.List[dict], use_auto_group_dict: bool = True
) -> dict:
    """
    It can group items separated by "__" into listed groups and separate them by selected primary-key values.
    Example: Primary key is ("a",)
    [
     {"a": 1, "b": 2, "c__a": 33, "c__b": 44},
     {"a": 1, "b": 2, "c__a": 55, "c__b": 66},
     {"a": 2, "b": 2, "c__a": 7, "c__b": 88},
     {"a": 2, "b": 2, "c__a": 77, "c__b": 99}
    ]
    ===>
    {
        '1': {'a': 1, 'b': 2, 'c': [{'a': 33, 'b': 44}, {'a': 55, 'b': 66}]},
        '2': {'a': 2, 'b': 2, 'c': [{'a': 7, 'b': 88}, {'a': 77, 'b': 99}]}
    }
    :param primary_key_names: Specify name of columns which you want to group by
    :param list_of_dicts: List of dicts. Usually SQL Select result.
    :param use_auto_group_dict: Should be used funtion auto_group_dict for each row of grouped keys into list?
    :return: Dict with joined repeated rows and listed dicts with group prefix stored under joined primary key.
    """
    pk_results = {}
    for item in list_of_dicts:
        primary_key = "-".join([str(item[pk_item]) for pk_item in primary_key_names])
        result = pk_results.get(primary_key, {})
        grouped_row = {}
        for key, value in item.items():
            parsed_key = re.findall(r"^([a-zA-Z0-9]+)__([^_]+.*)$", key)
            if parsed_key:
                group_key = parsed_key[0][0]
                subgroup_key = parsed_key[0][1]
                if group_key in grouped_row:
                    grouped_row[group_key][subgroup_key] = value
                else:
                    grouped_row[group_key] = {subgroup_key: value}
            else:
                if use_auto_group_dict:
                    result = auto_group_dict({key: value}, merge_with_dict=result)
                else:
                    result[key] = value
        for key, group_line in grouped_row.items():
            if use_auto_group_dict:
                group_line = auto_group_dict(group_line)
            if key in result:
                result[key].append(group_line)
            else:
                result[key] = [group_line]
        pk_results[primary_key] = result
    return pk_results


def auto_group_list(list_of_dicts: typing.List[dict], use_auto_group_dict: bool = True) -> dict:
    """
    It can group items separated by "__" into listed groups.
    Example: [
                 {"a": 1, "b": 2, "c__a": 3, "c__b": 4},
                 {"a": 1, "b": 2, "c__a": 5, "c__b": 6},
                 {"a": 1, "b": 2, "c__a": 7, "c__b": 8}
             ]
              ==>
            {"a": 1, "b": 2, "c": [
                {"a": 3, "b": 4},
                {"a": 5, "b": 6},
                {"a": 7, "b": 8}
              ]
            }
    :param list_of_dicts: List of dicts. Usually SQL Select result.
    :param use_auto_group_dict: Should be used funtion auto_group_dict for each row of grouped keys into list?
    :return: Dict with joined repeated rows and listed dicts with group prefix.
    """
    result = {}
    for item in list_of_dicts:
        grouped_row = {}
        for key, value in item.items():
            parsed_key = re.findall(r"^([a-zA-Z0-9]+)__([^_]+.*)$", key)
            if parsed_key:
                group_key = parsed_key[0][0]
                subgroup_key = parsed_key[0][1]
                if group_key in grouped_row:
                    grouped_row[group_key][subgroup_key] = value
                else:
                    grouped_row[group_key] = {subgroup_key: value}
            else:
                if use_auto_group_dict:
                    result = auto_group_dict({key: value}, merge_with_dict=result)
                else:
                    result[key] = value
        for key, group_line in grouped_row.items():
            if use_auto_group_dict:
                group_line = auto_group_dict(group_line)
            if key in result:
                result[key].append(group_line)
            else:
                result[key] = [group_line]
    return result


def auto_group_list_multi(list_of_dicts: typing.List[dict], use_auto_group_dict: bool = True) -> list:
    """
    It can group items separated by "__" into listed groups. It watch line change.
    If repeating key is changed, new line will of list is created.
    Example:
            [
                {"a": 1, "b": 2, "c__a": 3, "c__b": 4},
                {"a": 1, "b": 2, "c__a": 5, "c__b": 6},
                {"a": 1, "b": 2, "c__a": 7, "c__b": 8},
                {"a": 11, "b": 22, "c__a": 33, "c__b": 44},
                {"a": 11, "b": 22, "c__a": 55, "c__b": 66},
                {"a": 11, "b": 22, "c__a": 77, "c__b": 88},
            ]
            ==>
            [
                {"a": 1, "b": 2, "c": [
                        {"a": 3, "b": 4},
                        {"a": 5, "b": 6},
                        {"a": 7, "b": 8}
                    ]
                },
                {"a": 22, "b": 33, "c": [
                        {"a": 33, "b": 44},
                        {"a": 55, "b": 66},
                        {"a": 77, "b": 88}
                    ]
                }
            ]
    :param list_of_dicts: List of dicts. Usually SQL Select result.
    :param use_auto_group_dict: Should be used funtion auto_group_dict for each row of grouped keys into list?
    :return: Dict with joined repeated rows and listed dicts with group prefix.
    """
    results = []
    result_row: dict = {}
    last_row_identifier: list = None
    for item in list_of_dicts:
        row_identifier = []
        new_result_row = {}
        grouped_row = {}
        for key, value in item.items():
            parsed_key = re.findall(r"^([a-zA-Z0-9]+)__([^_]+.*)$", key)
            if parsed_key:
                group_key = parsed_key[0][0]
                subgroup_key = parsed_key[0][1]
                if group_key in grouped_row:
                    grouped_row[group_key][subgroup_key] = value
                else:
                    grouped_row[group_key] = {subgroup_key: value}
            else:
                if use_auto_group_dict:
                    new_result_row = auto_group_dict({key: value}, merge_with_dict=new_result_row)
                else:
                    new_result_row[key] = value
                row_identifier.append(value)

        if last_row_identifier != row_identifier:
            result_row = new_result_row
            results.append(result_row)
            last_row_identifier = row_identifier

        for key, group_line in grouped_row.items():
            if use_auto_group_dict:
                group_line = auto_group_dict(group_line)
            if key in result_row:
                result_row[key].append(group_line)
            else:
                result_row[key] = [group_line]

    return results


def auto_group_dict(dict_structure: dict, merge_with_dict: dict = None) -> dict:
    """
    It can group dict keys with same prefix under one dict key. Group keys are identified by group separator "___"
    Example: {
                "a": 1, "b": 2, "c___a": 3, "c___b___bb": 4, "c___b___cc": 5,
            }
            ===>
            {
                "a": 1, "b": 2, "c": {
                    "a": 3, "b": {
                        "bb": 4, "cc": 5
                    }
                }
            }
    :param dict_structure: Dict wit one level of depth. It ususaly goes from databese row select.
    :param merge_with_dict: Result will be added into this dict
    :return: Dict with posible N level structure
    """

    def set_value(_key_path, _value):
        _result = result
        key_path_len = len(_key_path)
        for index, _key in enumerate(_key_path):
            is_last = index == key_path_len - 1
            if _key not in _result:
                _result[_key] = {}
            if is_last:
                _result[_key] = _value
            _result = _result[_key]

    result = merge_with_dict if merge_with_dict else {}
    for key, value in dict_structure.items():
        key_path = key.split("___")
        if len(key_path) > 1:
            set_value(key_path, value)
        else:
            result[key] = value
    return result



import re


def dicts_into_list(dict_of_dicts: dict) -> list[dict]:
    """
    Converts dict of dicts into list structure.
    :param dict_of_dicts: Dict of dicts
    """
    return [item for _, item in dict_of_dicts.items()]


def sort_list_of_dicts(list_of_dicts: list[dict], sort_by: str) -> list[dict]:
    """
    Sort list of dicts by selected column.
    :param list_of_dicts: Input dict
    :param sort_by: Column name. To change direction of sort add +/- at the end of column name. Example: "id-"
    """
    sort_direction = "+"
    sort_order_match = re.match(r".*[+-]$", sort_by)
    if sort_order_match:
        sort_direction = sort_by[-1]
        sort_by = sort_by[:-1]
    result_list = sorted(list_of_dicts, key=lambda item_: item_[sort_by], reverse=(sort_direction == "-"))
    return result_list


def dict_pass(d: dict, attributes: list) -> dict:
    """
    Whitelist dictionary attributes.
    :param d: Dict instance
    :param attributes: Whitelisted attributes
    :return: New instance of filtered dict
    """
    return {key: value for key, value in d.items() if key in attributes}


def dict_filter(d: dict, attributes: list) -> dict:
    """
    Blacklist dictionary attributes.
    :param d: Dict instance
    :param attributes: Blacklisted attributes.
    :return: New instance of filtered dict
    """
    return {key: value for key, value in d.items() if key not in attributes}


def dict_swap(d: dict) -> dict:
    """
    Swap dict keys into values and values into keys.
    :param d: Dict to be swapped
    """
    return {str(value): key for key, value in d.items()}


def map_dict_into_list(d: dict, keys_map: list):
    """
    Maps dict keys into list positions based on keys_map defined in key map list.
    Example:
        d = {"a": "A value", "b": "B value", "c": "C value", "d": "D value"}
        map = ["c", "a", "b", "d"]
        ===>
        ["C value", "A value", "B value", "D value"]
    :param d: Dict to be mapped
    :param keys_map: List of keys order
    """
    result = [None] * len(keys_map)
    map_of_keys = dict_swap(dict(enumerate(keys_map)))

    for key, value in d.items():
        list_index = map_of_keys[key]
        result[list_index] = value
    return result



import re


def auto_group_list_into_tree_dict(tree_key_names: tuple, list_of_dicts: list[dict], include_key_names: bool = False) -> dict:
    """
    Group items based on values of given key names into dict tree structure recursively.
    Example: tree_key_names=("a", "b"), list_of_dicts=input1 = [
                {"a": "aa1", "b": "bb1", "c": 3, "d": 4},
                {"a": "aa1", "b": "bb2", "c": 33, "d": 44},
                {"a": "aa2", "b": "bb1", "c": 333, "d": 444},
                {"a": "aa2", "b": "bb2", "c": 3333, "d": 4444}
            ]
                ==>
            {
                'aa1': {
                            'bb1': [{'a': 'aa1', 'b': 'bb1', 'c': 3, 'd': 4}],
                            'bb2': [{'a': 'aa1', 'b': 'bb2', 'c': 33, 'd': 44}]
                },
                'aa2': {
                            'bb1': [{'a': 'aa2', 'b': 'bb1', 'c': 333, 'd': 444}],
                            'bb2': [{'a': 'aa2', 'b': 'bb2', 'c': 3333, 'd': 4444}]
                }
            }

    :param tree_key_names: Key names to create tree structure
    :param list_of_dicts: List of dicts to be transformed into tree
    :param include_key_names: Concat key name into key identifier separated by "|" char
    """
    def __get(dict_, key_value_):
        if isinstance(dict_.get(key_value_), list):
            return dict_[key_value_]
        dict_[key_value_] = []
        return dict_[key_value_]

    if not tree_key_names:
        return list_of_dicts

    result = {}
    key_name = tree_key_names[0]

    for item in list_of_dicts:
        result_key = "{}|{}".format(key_name, item.get(key_name)) if include_key_names else str(item.get(key_name))
        __get(result, result_key).append(item)

    if tree_key_names[1:]:
        for result_key, result_item in result.items():
            result[result_key] = auto_group_list_into_tree_dict(tree_key_names[1:], result_item, include_key_names)

    return result


def auto_group_list_into_tree_list(tree_key_names: tuple, list_of_dicts: list[dict]) -> list:
    """
    Group items based on values of given key names into list tree structure recursively.

        input1 = [
                        {"a": "aa1", "b": "bb1", "c": 3, "d": 4},
                        {"a": "aa1", "b": "bb2", "c": 33, "d": 44},
                        {"a": "aa2", "b": "bb1", "c": 333, "d": 444},
                        {"a": "aa2", "b": "bb2", "c": 3333, "d": 4444}
        ]

    ==>
        [
            {'a': 'aa1', 'items': [
                {'b': 'bb1', 'items': [
                    {'a': 'aa1', 'b': 'bb1', 'c': 3, 'd': 4}
                ]},
                {'b': 'bb2', 'items': [
                    {'a': 'aa1', 'b': 'bb2', 'c': 33, 'd': 44}]}]
             },
            {'a': 'aa2', 'items': [
                {'b': 'bb1', 'items': [
                    {'a': 'aa2', 'b': 'bb1', 'c': 333, 'd': 444}
                ]},
                {'b': 'bb2', 'items': [
                    {'a': 'aa2', 'b': 'bb2', 'c': 3333, 'd': 4444}]
                 }]
             }
        ]
    :param tree_key_names: Key names to create tree structure. By adding +/- at the end of key name order of items (ASC or DESC) could be driven.
    :param list_of_dicts: List of dicts to be transformed into tree
    """
    def __get(items_, key_, key_value_):
        for item_ in items_:
            if item_.get(key_) == key_value_:
                return item_["items"]
        items_.append({
            key_: key_value_,
            "items": [],
        })
        return items_[-1]["items"]

    if not tree_key_names:
        return list_of_dicts

    result = []
    key_name = tree_key_names[0]
    sort_order = None
    sort_order_match = re.match(r".*[+-]$", key_name)
    if sort_order_match:
        sort_order = key_name[-1]
        key_name = key_name[:-1]

    for item in list_of_dicts:
        result_key = item.get(key_name)
        __get(result, key_name, result_key).append(item)

    if tree_key_names[1:]:
        for result_index, result_item in enumerate(result):
            result[result_index]["items"] = auto_group_list_into_tree_list(tree_key_names[1:], result_item["items"])

        if sort_order:
            result = sorted(result, key=lambda item_: item_[key_name], reverse=(sort_order == "-"))

    return result



#!/usr/bin/python3
from setuptools import find_packages, setup


def get_file_content_as_list(file_path: str) -> list:
    with open(file_path, "r", encoding="utf-8") as file:
        content = [line.strip() for line in file]
    return content


def get_file_content(file_path: str) -> str:
    with open(file_path, "r", encoding="utf-8") as file:
        content = file.read()
    return content


packages = find_packages(exclude=["tests"])
print(f"*****\n*****\nFound packages: {packages}")
VERSION = get_file_content("VERSION")
INSTALL_REQUIRES = get_file_content_as_list("requirements.txt")
DOCUMENTATION_MD = get_file_content("README.md")

setup(
    name="auto-group",
    version=VERSION,
    author="Ales Adamek, Filip Cima",
    author_email="alda78@seznam.cz",
    description="Tools for creating tree structures from flat list of dicts",
    long_description=DOCUMENTATION_MD,
    long_description_content_type="text/markdown",
    url="https://gitlab.com/alda78/auto-group",
    packages=packages,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
    install_requires=INSTALL_REQUIRES,
    include_package_data=True,  # MANIFEST.in
    zip_safe=False,  # aby se spravne vycitala statika pridana pomoci MANIFEST.in
)
